/* Copyright 2010, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

//`define TESTBENCH

`ifdef TESTBENCH
`timescale 1 ns / 1 ps
`endif

module weim_wishbone_bridge (
  wbm_clk_i,
  wbm_adr_o,
  wbm_dat_i,
  wbm_ack_i,
  wbm_cyc_o,
  wbm_stb_o,
  wbm_we_o,
  wbm_dat_o,
  wbm_sel_o,

  rst_i,
  eim_d_i,
  eim_d_o,
  eim_d_oe_o,
  eim_bclk_i,
  eim_lba_i,
  eim_rd_n_i,
  eim_wr_n_i,
  eim_wait_n_o,
  eim_cs0_n_i,
  eim_cs1_n_i,
  eim_byte0_n_i,
  eim_byte1_n_i,

  ts4800_decodes_o
);

input wbm_clk_i;
output [31:0] wbm_adr_o;
input [15:0] wbm_dat_i /* synthesis use_ioff=0 */;
input wbm_ack_i;
output wbm_cyc_o;
output wbm_stb_o;
output wbm_we_o;
output [15:0] wbm_dat_o;
output [1:0] wbm_sel_o;

input rst_i;
input [15:0] eim_d_i;
output [15:0] eim_d_o;
output eim_d_oe_o;
input eim_bclk_i;
input eim_lba_i;
input eim_rd_n_i;
input eim_wr_n_i;
output eim_wait_n_o;
input eim_cs0_n_i; /* 16 bit address space */
input eim_cs1_n_i; /* 8 bit address space */
input eim_byte0_n_i;
input eim_byte1_n_i;

output [16:0] ts4800_decodes_o;

reg [2:0] state;
reg [1:0] wbm_sel;
reg [31:0] wbm_adr;
reg [15:0] wbm_dat;
reg wbm_we, wbm_cyc, wbm_stb, wbm_err;
reg eim_rd, eim_wr, eim_lba, eim_byte0, eim_byte1, eim_dat_oe;
reg [15:0] eim_dat;
reg [15:0] eim_mux;
reg cs, cs0, cs1, rd, wr;
wire [31:0] wbm_odat;
wire wbm_iack;
assign eim_d_oe_o = eim_dat_oe && (!eim_cs0_n_i || !eim_cs1_n_i);
assign eim_d_o = eim_dat;

wire [31:0] wbm_dat16_i = {16'd0, wbm_dat_i[15:0]};
wire [31:0] wbm_dat16_o;
assign wbm_dat_o = wbm_dat16_o[15:0];
wire [3:0] wbm_sel16_o;
assign wbm_sel_o = wbm_sel16_o[1:0];
wb_writebuffer wrfifo(
  .wb_clk_i(wbm_clk_i),
  .wb_rst_i(rst_i),
  .wb_adr_i(wbm_adr),
  .wb_dat_i({16'd0, wbm_dat}),
  .wb_dat_o(wbm_odat),
  .wb_cyc_i(wbm_cyc),
  .wb_stb_i(wbm_stb),
  .wb_we_i(wbm_we),
  .wb_ack_o(wbm_iack),
  .wb_sel_i({2'd0, wbm_sel}),

  .wbm_adr_o(wbm_adr_o),
  .wbm_dat_i(wbm_dat16_i),
  .wbm_dat_o(wbm_dat16_o),
  .wbm_cyc_o(wbm_cyc_o),
  .wbm_stb_o(wbm_stb_o),
  .wbm_we_o(wbm_we_o),
  .wbm_ack_i(wbm_ack_i),
  .wbm_sel_o(wbm_sel16_o),

  .flushreq_i(1'b0)
);

reg [31:0] wbm_adr_q;
reg [1:0] wbm_sel_q;
always @(posedge eim_lba_i or posedge rst_i) begin
  if (rst_i) begin
    wbm_adr_q <= 32'd0;
    wbm_sel_q <= 2'd0;
  end else begin
    wbm_adr_q[15:0] <= eim_d_i;
    wbm_sel_q <= {!eim_byte1_n_i, !eim_byte0_n_i};
  end
end

always @(posedge eim_wr_n_i or posedge rst_i) begin
  if (rst_i) wbm_dat <= 16'd0;
  else wbm_dat <= eim_d_i;
end

reg eim_wait, stall;
assign eim_wait_n_o = !eim_wait;
always @(*) begin
  eim_wait = 1'b0;

  if ((!eim_cs0_n_i || !eim_cs1_n_i) && (!eim_rd_n_i || !eim_wr_n_i)) begin
    eim_wait = stall;
  end

end

wire [17:0] adr = cs0 ? {1'b0, wbm_adr_q[15:0], 1'b0} : {2'b10, wbm_adr_q[15:0]};
wire [17:0] decodes = adr[16] ? (2 << adr[15:12]) : 1;
reg [17:0] ts4800_decodes;
assign ts4800_decodes_o = ts4800_decodes;

always @(posedge wbm_clk_i or posedge rst_i) begin
  if (rst_i) begin
    state <= 3'd0;
    cs0 <= 1'b0;
    cs1 <= 1'b0;
    cs <= 1'b0;
    rd <= 1'b0;
    wr <= 1'b0;
    eim_dat_oe <= 1'b0;
    wbm_cyc <= 1'b0;
    wbm_stb <= 1'b0;
    wbm_we <= 1'b0;
    eim_dat <= 16'd0;
    wbm_adr <= 32'd0;
    wbm_sel <= 2'd0;
    stall <= 1'b1;
    ts4800_decodes <= 17'd0;
  end else begin
    cs0 <= !eim_cs0_n_i;
    cs1 <= !eim_cs1_n_i;
    rd <= !eim_rd_n_i;
    wr <= !eim_wr_n_i;

    case (state)
    3'd0: begin
      if ((cs0 || cs1) && rd) begin
        cs <= cs1 ? 1'b1 : 1'b0;
        state <= 3'd1;
        wbm_adr <= adr;
        ts4800_decodes <= decodes;
        wbm_sel <= cs0 ? wbm_sel_q : 2'b01;
        wbm_cyc <= 1'b1;
        wbm_stb <= 1'b1;
        wbm_we <= 1'b0;
        eim_dat_oe <= 1'b1;
      end else if ((cs0 || cs1) && wr) begin
        wbm_adr <= adr;
        ts4800_decodes <= decodes;
        wbm_sel <= cs0 ? wbm_sel_q : 2'b01;
        cs <= cs1 ? 1'b1 : 1'b0;
        state <= 3'd3; /* wait for write to deassert before starting wb */
        stall <= 1'b0;
      end
    end
    3'd1: if (wbm_iack) begin
      wbm_cyc <= 1'b0;
      wbm_stb <= 1'b0;
      eim_dat <= wbm_odat[15:0];
      state <= 3'd2; /* cycle is done, wait for cs to deassert */
      stall <= 1'b0;
    end
    3'd2: if (!rd) begin
      state <= 3'd0; /* bus cycle is complete */
      eim_dat_oe <= 1'b0;
      stall <= 1'b1;
    end
    3'd3: if (!wr) begin
      wbm_cyc <= 1'b1;
      wbm_stb <= 1'b1;
      wbm_we <= 1'b1;
      state <= 3'd4;
      stall <= 1'b1;
    end
    3'd4: begin
      if (wbm_iack) begin
        wbm_cyc <= 1'b0;
        wbm_stb <= 1'b0;
        state <= 3'd0;
      end
    end
    endcase

  end
end


endmodule

`ifdef TESTBENCH
module test();

reg clk, rst;
wire wbm_cyc, wbm_stb, wbm_we, eim_d_oe;
wire [15:0] wbm_odat, eim_d_o;
reg [15:0] wbm_idat, eim_d;
wire [31:0] wbm_adr;
wire [1:0] wbm_sel;
reg eim_lba, eim_rd, eim_wr, eim_cs0, eim_cs1, wbm_ack;
wire eim_wait;

weimbridge uut(
  .wbm_clk_i(clk),
  .wbm_adr_o(wbm_adr),
  .wbm_dat_i(wbm_idat),
  .wbm_ack_i(wbm_ack),
  .wbm_cyc_o(wbm_cyc),
  .wbm_stb_o(wbm_stb),
  .wbm_we_o(wbm_we),
  .wbm_dat_o(wbm_odat),
  .wbm_sel_o(wbm_sel),

  .rst_i(rst),
  .eim_d_i(eim_d),
  .eim_d_o(eim_d_o),
  .eim_d_oe_o(eim_d_oe),
  .eim_lba_i(eim_lba),
  .eim_rd_n_i(eim_rd),
  .eim_wr_n_i(eim_wr),
  .eim_wait_n_o(eim_wait),
  .eim_cs0_n_i(eim_cs0),
  .eim_cs1_n_i(eim_cs1),
  .eim_byte0_n_i(1'b1),
  .eim_byte1_n_i(1'b1)
);

always #5 clk = !clk;

reg [3:0] delay, delaycnt;
always @(posedge clk or posedge rst) begin
  if (rst) begin
    wbm_idat <= 16'd0;
    wbm_ack <= 1'b0;
    delaycnt <= 4'd0;
  end else begin
    wbm_ack <= 1'b0;
    if (wbm_cyc && wbm_stb) delaycnt <= delaycnt + 1'b1;
    if (wbm_cyc && wbm_stb && !wbm_ack && delaycnt == delay) begin
      wbm_ack <= 1'b1;
      delaycnt <= 4'd0;
      if (wbm_we) $display("detected WISHBONE write-- dat:%x adr:%x", wbm_odat, wbm_adr);
      else $display("detected WISHBONE read-- adr:%x", wbm_adr);
    end
    wbm_idat <= wbm_idat + 1;
  end
end

initial begin
  $dumpfile("dump.vcd");
  $dumpvars(3, test);

  eim_cs0 = 1; eim_cs1 = 1; eim_rd = 1; eim_wr = 1; eim_lba = 1;
  eim_d = 16'hxxxx; wbm_idat = 32'd0; 
  rst = 1; clk = 0;
  delay = 4'd4;
  #100;
  rst = 0;
  #10;
  @(posedge clk);

  /* Write bus cycle */
  eim_cs0 = 1'b0; eim_lba = 1'b0; eim_wr = 1'b0;
  #23;
  eim_d = 16'habcd;
  #23;
  eim_lba = 1'b1;
  #23;
  eim_d = 16'h0123;
  eim_wr = 1'b0;
  #75;
  eim_wr = 1'b1;
  #3;
  eim_cs0 = 1'b1;
  #50;

  /* Write bus cycle */
  eim_cs0 = 1'b0; eim_lba = 1'b0; eim_wr = 1'b0;
  #23;
  eim_d = 16'h4568;
  #23;
  eim_lba = 1'b1;
  #23;
  eim_d = 16'h3210;
  eim_wr = 1'b0;
  #75;
  eim_wr = 1'b1;
  #3;
  eim_cs0 = 1'b1;
  #100;

  /* Read bus cycle */
  eim_cs0 = 1'b0; eim_lba = 1'b0; 
  #23;
  eim_d = 16'habce;
  #23;
  eim_lba = 1'b1;
  #23;
  eim_rd = 1'b0;
  #75;
  $display("EIM read data: %x", eim_d_o);
  eim_rd = 1'b1;
  #3;
  eim_cs0 = 1'b1;
  #100;


  /* Read 8bit bus cycle */
  eim_cs1 = 1'b0; eim_lba = 1'b0; 
  #23;
  eim_d = 16'habce;
  #23;
  eim_lba = 1'b1;
  #23;
  eim_rd = 1'b0;
  #75;
  $display("EIM read data: %x", eim_d_o);
  eim_rd = 1'b1;
  #3;
  eim_cs1 = 1'b1;
  #100;


  $finish;
end


endmodule
`endif
